@@ -1,6 +1,6 @@
-// $Id: yascreen.c,v 1.85 2020/10/01 03:48:48 bbonev Exp $
+// $Id: yascreen.c,v 1.99 2023/08/02 17:32:12 bbonev Exp $
 //
-// Copyright © 2015-2020 Boian Bonev (bbonev@ipacct.com) {{{
+// Copyright © 2015-2023 Boian Bonev (bbonev@ipacct.com) {{{
 //
 // SPDX-License-Identifer: LGPL-3.0-or-later
 //
@@ -44,7 +44,7 @@
 #define PSIZE (sizeof(char *))
 // step to allocate key buffer
 #define KEYSTEP (4096/sizeof(int))
-// default timeout before escape is returned
+// default timeout in milliseconds before escape is returned
 #define YAS_DEFAULT_ESCTO 300
 
 // check if a given value is a valid simple color value
@@ -96,6 +96,7 @@ typedef enum { // ansi sequence state machine
 	ST_ESC_SQ, // escape [ sequence
 	ST_ESC_SQ_D, // escape [ digit sequence
 	ST_ESC_O, // escape O sequence
+	ST_ESC_ESC, // escape escape sequence
 } yas_k_state;
 
 typedef enum { // telnet sequence state machine
@@ -132,7 +133,7 @@ struct _yascreen {
 	cell *scr; // screen state
 	struct termios *tsstack; // saved terminal state
 	int tssize; // number of items in the stack
-	int escto; // single ESC key timeout
+	int escto; // single ESC key timeout (in milliseconds)
 	int keysize; // saved key storage size
 	int keycnt; // saved key count
 	int *keys; // saved key array
@@ -153,6 +154,7 @@ struct _yascreen {
 	uint8_t haveansi:1; // we do have a reported screen size from ansi sequence
 	uint8_t havenaws:1; // we do have a reported screen size from telent naws
 	uint8_t istelnet:1; // do process telnet sequences
+	uint8_t isunicode:1; // do process unicode sequences
 	uint8_t cursor:1; // last cursor state
 	uint8_t redraw:1; // flag to redraw from scratch
 	uint8_t lineflush:1; // always flush after line operations
@@ -161,6 +163,28 @@ struct _yascreen {
 	uint8_t outb[256]; // buffered output
 	uint16_t outp; // position in outb
 };
+
+// helpers for versioned symbols
+#if YASCREEN_VERSIONED // implementation always follows the attribute
+#if defined __GNUC__ && __GNUC__ >= 10 && !defined(__clang__)
+#define symver_o(impl,sym,ver) __attribute__((symver(#sym"@"#ver)))
+#define symver_d(impl,sym,ver) __attribute__((symver(#sym"@@"#ver)))
+#else
+#define symver_o(impl,sym,ver) asm(".symver "#impl","#sym"@"#ver);
+#define symver_d(impl,sym,ver) asm(".symver "#impl","#sym"@@"#ver);
+#endif
+#define VV(foo,ver) foo##ver
+#define V(foo,ver) VV(foo,ver)
+#define V193 _193
+#else
+#define symver_o(impl,sym,ver)
+#define symver_d(impl,sym,ver)
+#define V(foo,ver) foo
+#define V193
+#endif
+#define symver_V(impl,sym,ver) symver_d(impl,sym,ver)
+#define YASCREEN_193 YASCREEN_1.93
+
 // }}}
 
 static inline int64_t mytime() { // {{{
@@ -290,7 +314,7 @@ inline void *yascreen_get_hint_p(yascreen *s) { // {{{
 	return s->phint;
 } // }}}
 
-static char myver[]="\0Yet another screen library (https://github.com/bbonev/yascreen) $Revision: 1.85 $\n\n"; // {{{
+static char myver[]="\0Yet another screen library (https://github.com/bbonev/yascreen) $Revision: 1.99 $\n\n"; // {{{
 // }}}
 
 inline const char *yascreen_ver(void) { // {{{
@@ -321,14 +345,7 @@ inline yascreen *yascreen_init(int sx,int sy) { // {{{
 	if (!s)
 		return NULL;
 
-	s->state=ST_NORM;
-	s->tstate=T_NORM;
-	s->ustate=U_NORM;
-	s->escto=YAS_DEFAULT_ESCTO;
-	s->cursor=1;
-	s->cursorx=0;
-	s->cursory=0;
-	if (!s->outcb&&isatty(STDOUT_FILENO)) { // output is a terminal
+	if (/*!s->outcb&&*/isatty(STDOUT_FILENO)) { // output is a terminal
 		s->tsstack=(struct termios *)calloc(1,sizeof(struct termios));
 		if (!s->tsstack) {
 			free(s);
@@ -355,6 +372,39 @@ inline yascreen *yascreen_init(int sx,int sy) { // {{{
 	}
 	s->sx=sx;
 	s->sy=sy;
+	// s->outcb is already NULL
+	// s->mem is initialized below
+	// s->scr is initialized below
+	// s->tstack is initialized above
+	// s->tssize is initialized above
+	s->escto=YAS_DEFAULT_ESCTO;
+	// s->keysize is initialized below
+	// s->keycnt is already zero
+	// s->keys is initialized below
+	// s->ansibuf is already zeroes
+	// s->ansipos is already zero
+	// s->sosnbuf is already zeroes
+	// s->sosnpos is already zero
+	// s->utf is already zeroes
+	// s->escts is already zero
+	s->state=ST_NORM;
+	s->tstate=T_NORM;
+	s->ustate=U_NORM;
+	// s->cursorx is already zero
+	// s->cursory is already zero
+	// s->scrx is already zero
+	// s->scry is already zero
+	// s->haveansi is already zero
+	// s->havenaws is already zero
+	// s->istelnet is already zero
+	s->isunicode=1; // previous versions were unicode only
+	s->cursor=1; // cursor is visible by default
+	s->redraw=1; // leave scr empty, so that on first refresh everything is redrawn
+	s->lineflush=1; // be compatible with earlier versions that worked without output buffering normally a recent client will set this to 0 and use explicit flush
+	// s->hint is already zero
+	// s->phint is already NULL
+	// s->outb is already zeroes
+	// s->outp is already zero
 
 	s->keys=(int *)calloc(KEYSTEP,sizeof(int));
 	if (!s->keys) {
@@ -381,11 +431,6 @@ inline yascreen *yascreen_init(int sx,int sy) { // {{{
 
 	for (i=0;i<sx*sy;i++)
 		strncpy(s->mem[i].d," ",sizeof s->mem[i].d);
-	// be compatible with earlier versions that worked without output buffering
-	// normally a recent client will set this to 0 and use explicit flush
-	s->lineflush=1;
-	// leave scr empty, so that on first refresh everything is redrawn
-	s->redraw=1;
 	return s;
 } // }}}
 
@@ -404,6 +449,13 @@ inline void yascreen_set_telnet(yascreen *s,int on) { // {{{
 	s->istelnet=!!on;
 } // }}}
 
+inline void yascreen_set_unicode(yascreen *s,int on) { // {{{
+	if (!s)
+		return;
+	s->isunicode=!!on;
+	s->keycnt=0; // flush input buffer - it may not be verified unicode
+} // }}}
+
 inline void yascreen_init_telnet(yascreen *s) { // {{{
 	if (!s)
 		return;
@@ -529,11 +581,11 @@ inline void yascreen_update_attr(yascreen *s,uint32_t oattr,uint32_t nattr) { //
 
 	if (oattr==0xffffffff) {
 		oattr=~nattr; // force setting all
-		outs(s,ESC"0m");
+		outs(s,ESC"[0m");
 	}
 
 	if ((oattr&YAS_BOLD)!=(nattr&YAS_BOLD))
-		outs(s,(nattr&YAS_BOLD)?ESC"[1m":ESC"[21m");
+		outs(s,(nattr&YAS_BOLD)?ESC"[1m":ESC"[22m");
 	if ((oattr&YAS_ITALIC)!=(nattr&YAS_ITALIC))
 		outs(s,(nattr&YAS_ITALIC)?ESC"[3m":ESC"[23m");
 	if ((oattr&YAS_UNDERL)!=(nattr&YAS_UNDERL))
@@ -619,7 +671,7 @@ static inline int yascreen_update_range(yascreen *s,int y1,int y2) { // {{{
 						outs(s,(s->mem[j*s->sx+i-cnt].style&YAS_STORAGE)?s->mem[j*s->sx+i-cnt].p:s->mem[j*s->sx+i-cnt].d);
 						cnt--;
 					}
-					cnt=0; // cycle above leaves cnt at -1
+					cnt=0; // loop above leaves cnt at -1
 				} else
 					cnt++;
 			}
@@ -1388,14 +1440,43 @@ inline int yascreen_y(yascreen *s) { // {{{
 } // }}}
 
 inline void yascreen_ckto(yascreen *s) { // {{{
+	int64_t now;
+
 	if (!s)
 		return;
 
-	if (s->state==ST_ESC&&s->ansipos==1&&s->ansibuf[0]==YAS_K_ESC&&s->escto&&s->escts+s->escto<mytime()) {
+	now=mytime();
+
+	if (s->state==ST_ESC&&s->ansipos==1&&s->escto&&s->escts+s->escto<now) {
 		s->ansipos=0;
 		s->state=ST_NORM;
 		yascreen_pushch(s,YAS_K_ESC);
 	}
+	if (s->state==ST_ESC_ESC&&s->ansipos==2&&s->escto&&s->escts+s->escto<now) {
+		s->ansipos=0;
+		s->state=ST_NORM;
+		yascreen_pushch(s,YAS_K_ESC); // return a double escape as a signle one, is that the proper way?
+	}
+	if ((s->state==ST_ESC||s->state==ST_ESC_ESC)&&s->escto&&s->escts+s->escto<now) {
+		s->ansipos=0;
+		s->state=ST_NORM; // ignore slow escape sequences
+	}
+} // }}}
+
+inline uint64_t yascreen_willto(yascreen *s) { // {{{
+	int64_t now;
+
+	if (!s)
+		return 0;
+
+	if (s->state!=ST_ESC||s->ansipos!=1||s->ansibuf[0]!=YAS_K_ESC||!s->escto)
+		return 0; // no timeout pending
+
+	now=mytime();
+	if (s->escts+s->escto<=now)
+		return 1; // timeout already expired
+	else
+		return s->escts+s->escto-now;
 } // }}}
 
 static inline int yascreen_feed_telnet(yascreen *s,unsigned char c) { // {{{
@@ -1482,471 +1563,20 @@ static inline int yascreen_feed_telnet(yascreen *s,unsigned char c) { // {{{
 	return TELNET_NOOP;
 } // }}}
 
-inline void yascreen_feed(yascreen *s,unsigned char c) { // {{{
-	if (!s)
-		return;
+#if YASCREEN_VERSIONED
+inline void yascreen_feed_193(yascreen *s,unsigned char c);
+inline void yascreen_feed_179(yascreen *s,unsigned char c);
+#endif
 
-	yascreen_ckto(s);
-	if (s->istelnet) { // process telnet codes
-		int tc=yascreen_feed_telnet(s,c);
-
-		switch (tc) {
-			case 0x00 ... 0xff: // normal character
-				c=(unsigned char)tc;
-				break;
-			default:
-			case TELNET_NOOP: // byte is eaten w/o valid input
-				return;
-			case TELNET_SIZE: // notify about screen size change
-				yascreen_pushch(s,YAS_TELNET_SIZE);
-				return;
-		}
-	}
-
-	switch (s->state) {
-		case ST_ENTER:
-			if (c=='\n'||c==0) // ignore LF or NUL after CR
-				break;
-			s->state=ST_NORM;
-			// fall through
-		case ST_NORM:
-			if (c==YAS_K_ESC) { // handle esc sequences
-				s->escts=mytime();
-				s->ansipos=1;
-				s->ansibuf[0]=c;
-				s->state=ST_ESC;
-			} else { // handle standard keys
-				if (c=='\r') // shift state to ST_ENTER to eat optional LF/NUL after CR
-					s->state=ST_ENTER;
-				switch (s->ustate) {
-					case U_NORM:
-						if (c&0x80) {
-							if ((c&0xc0)==0x80) // unexpected continuation byte - ignore
-								break;
-						startbyte:
-							if ((c&0xe0)==0xc0) { // 2 byte seq
-								s->utf[0]=c;
-								s->ustate=U_L2C1;
-								break;
-							}
-							if ((c&0xf0)==0xe0) { // 3 byte seq
-								s->utf[0]=c;
-								s->ustate=U_L3C1;
-								break;
-							}
-							if ((c&0xf8)==0xf0) { // 4 byte seq
-								s->utf[0]=c;
-								s->ustate=U_L4C1;
-								break;
-							}
-							if ((c&0xfc)==0xf8) { // 5 byte seq
-								//s->utf[0]=c;
-								s->ustate=U_L5C1;
-								break;
-							}
-							if ((c&0xfe)==0xfc) { // 6 byte seq
-								//s->utf[0]=c;
-								s->ustate=U_L6C1;
-								break;
-							}
-							// pass 0xff and 0xfe - violates rfc
-							yascreen_pushch(s,c);
-							s->ustate=U_NORM; // in case we come from unexpected start byte
-						} else
-							yascreen_pushch(s,c);
-						break;
-					case U_L2C1:
-						if ((c&0xc0)==0x80) { // continuation byte
-							yascreen_pushch(s,s->utf[0]);
-							yascreen_pushch(s,c);
-							s->ustate=U_NORM;
-							break;
-						}
-						if (c&0x80) // start another sequence
-							goto startbyte;
-						s->ustate=U_NORM; // normal byte kills current sequence and is processed
-						yascreen_pushch(s,c);
-						break;
-					case U_L3C1:
-						if ((c&0xc0)==0x80) { // continuation byte
-							s->utf[1]=c;
-							s->ustate=U_L3C2;
-							break;
-						}
-						if (c&0x80) // start another sequence
-							goto startbyte;
-						s->ustate=U_NORM; // normal byte kills current sequence and is processed
-						yascreen_pushch(s,c);
-						break;
-					case U_L3C2:
-						if ((c&0xc0)==0x80) { // continuation byte
-							yascreen_pushch(s,s->utf[0]);
-							yascreen_pushch(s,s->utf[1]);
-							yascreen_pushch(s,c);
-							s->ustate=U_NORM;
-							break;
-						}
-						if (c&0x80) // start another sequence
-							goto startbyte;
-						s->ustate=U_NORM; // normal byte kills current sequence and is processed
-						yascreen_pushch(s,c);
-						break;
-					case U_L4C1:
-						if ((c&0xc0)==0x80) { // continuation byte
-							s->utf[1]=c;
-							s->ustate=U_L4C2;
-							break;
-						}
-						if (c&0x80) // start another sequence
-							goto startbyte;
-						s->ustate=U_NORM; // normal byte kills current sequence and is processed
-						yascreen_pushch(s,c);
-						break;
-					case U_L4C2:
-						if ((c&0xc0)==0x80) { // continuation byte
-							s->utf[2]=c;
-							s->ustate=U_L4C3;
-							break;
-						}
-						if (c&0x80) // start another sequence
-							goto startbyte;
-						s->ustate=U_NORM; // normal byte kills current sequence and is processed
-						yascreen_pushch(s,c);
-						break;
-					case U_L4C3:
-						if ((c&0xc0)==0x80) { // continuation byte
-							yascreen_pushch(s,s->utf[0]);
-							yascreen_pushch(s,s->utf[1]);
-							yascreen_pushch(s,s->utf[2]);
-							yascreen_pushch(s,c);
-							s->ustate=U_NORM;
-							break;
-						}
-						if (c&0x80) // start another sequence
-							goto startbyte;
-						s->ustate=U_NORM; // normal byte kills current sequence and is processed
-						yascreen_pushch(s,c);
-						break;
-					case U_L5C1:
-						if ((c&0xc0)==0x80) { // continuation byte
-							//s->utf[1]=c;
-							s->ustate=U_L5C2;
-							break;
-						}
-						if (c&0x80) // start another sequence
-							goto startbyte;
-						s->ustate=U_NORM; // normal byte kills current sequence and is processed
-						yascreen_pushch(s,c);
-						break;
-					case U_L5C2:
-						if ((c&0xc0)==0x80) { // continuation byte
-							//s->utf[2]=c;
-							s->ustate=U_L5C3;
-							break;
-						}
-						if (c&0x80) // start another sequence
-							goto startbyte;
-						s->ustate=U_NORM; // normal byte kills current sequence and is processed
-						yascreen_pushch(s,c);
-						break;
-					case U_L5C3:
-						if ((c&0xc0)==0x80) { // continuation byte
-							//s->utf[3]=c;
-							s->ustate=U_L5C4;
-							break;
-						}
-						if (c&0x80) // start another sequence
-							goto startbyte;
-						s->ustate=U_NORM; // normal byte kills current sequence and is processed
-						yascreen_pushch(s,c);
-						break;
-					case U_L5C4:
-						if ((c&0xc0)==0x80) { // continuation byte
-							//yascreen_pushch(s,s->utf[0]); // sequence is parsed but ignored
-							//yascreen_pushch(s,s->utf[1]);
-							//yascreen_pushch(s,s->utf[2]);
-							//yascreen_pushch(s,s->utf[3]);
-							//yascreen_pushch(s,c);
-							s->ustate=U_NORM;
-							break;
-						}
-						if (c&0x80) // start another sequence
-							goto startbyte;
-						s->ustate=U_NORM; // normal byte kills current sequence and is processed
-						yascreen_pushch(s,c);
-						break;
-					case U_L6C1:
-						if ((c&0xc0)==0x80) { // continuation byte
-							//s->utf[1]=c;
-							s->ustate=U_L6C2;
-							break;
-						}
-						if (c&0x80) // start another sequence
-							goto startbyte;
-						s->ustate=U_NORM; // normal byte kills current sequence and is processed
-						yascreen_pushch(s,c);
-						break;
-					case U_L6C2:
-						if ((c&0xc0)==0x80) { // continuation byte
-							//s->utf[2]=c;
-							s->ustate=U_L6C3;
-							break;
-						}
-						if (c&0x80) // start another sequence
-							goto startbyte;
-						s->ustate=U_NORM; // normal byte kills current sequence and is processed
-						yascreen_pushch(s,c);
-						break;
-					case U_L6C3:
-						if ((c&0xc0)==0x80) { // continuation byte
-							//s->utf[3]=c;
-							s->ustate=U_L6C4;
-							break;
-						}
-						if (c&0x80) // start another sequence
-							goto startbyte;
-						s->ustate=U_NORM; // normal byte kills current sequence and is processed
-						yascreen_pushch(s,c);
-						break;
-					case U_L6C4:
-						if ((c&0xc0)==0x80) { // continuation byte
-							//s->utf[3]=c;
-							s->ustate=U_L6C5;
-							break;
-						}
-						if (c&0x80) // start another sequence
-							goto startbyte;
-						s->ustate=U_NORM; // normal byte kills current sequence and is processed
-						yascreen_pushch(s,c);
-						break;
-					case U_L6C5:
-						if ((c&0xc0)==0x80) { // continuation byte
-							//yascreen_pushch(s,s->utf[0]); // sequence is parsed but ignored
-							//yascreen_pushch(s,s->utf[1]);
-							//yascreen_pushch(s,s->utf[2]);
-							//yascreen_pushch(s,s->utf[3]);
-							//yascreen_pushch(s,s->utf[4]);
-							//yascreen_pushch(s,c);
-							s->ustate=U_NORM;
-							break;
-						}
-						if (c&0x80) // start another sequence
-							goto startbyte;
-						s->ustate=U_NORM; // normal byte kills current sequence and is processed
-						yascreen_pushch(s,c);
-						break;
-				}
-			}
-			break;
-		case ST_ESC:
-			switch (c) {
-				case '`':
-				case '-':
-				case '=':
-				case 0x7f:
-				case '~':
-				case '!':
-				case '@':
-				case '#':
-				case '$':
-				case '%':
-				case '^':
-				case '&':
-				case '*':
-				case '(':
-				case ')':
-				case '_':
-				case '+':
-				case ':':
-				case ';':
-				case '"':
-				case '\'':
-				case '{':
-				case '}':
-				case '|':
-				case '\\':
-				case ',':
-				case '.':
-				case '/':
-				case '<':
-				case '>':
-				case '?':
-				case '0'...'9':
-				case 'a'...'z':
-					yascreen_pushch(s,YAS_K_ALT(c));
-					s->state=ST_NORM;
-					break;
-				case '[':
-					s->ansibuf[s->ansipos++]=c;
-					s->state=ST_ESC_SQ;
-					break;
-				case 'O':
-					s->ansibuf[s->ansipos++]=c;
-					s->state=ST_ESC_O;
-					break;
-				default: // ignore unknown sequence
-					s->state=ST_NORM;
-					break;
-			}
-			break;
-		case ST_ESC_SQ:
-			switch (c) {
-				case 'A': // up
-					yascreen_pushch(s,YAS_K_UP);
-					s->state=ST_NORM;
-					break;
-				case 'B': // down
-					yascreen_pushch(s,YAS_K_DOWN);
-					s->state=ST_NORM;
-					break;
-				case 'C': // right
-					yascreen_pushch(s,YAS_K_RIGHT);
-					s->state=ST_NORM;
-					break;
-				case 'D': // left
-					yascreen_pushch(s,YAS_K_LEFT);
-					s->state=ST_NORM;
-					break;
-				case 'H': // home
-					yascreen_pushch(s,YAS_K_HOME);
-					s->state=ST_NORM;
-					break;
-				case 'F': // end
-					yascreen_pushch(s,YAS_K_END);
-					s->state=ST_NORM;
-					break;
-				case '0'...'9':
-					s->state=ST_ESC_SQ_D;
-					s->ansibuf[s->ansipos++]=c;
-					break;
-				default: // ignore unknown sequence
-					s->state=ST_NORM;
-					break;
-			}
-			break;
-		case ST_ESC_SQ_D:
-			if (s->ansipos>=sizeof s->ansibuf) { // buffer overrun, ignore the sequence
-				s->state=ST_NORM;
-				break;
-			}
-			s->ansibuf[s->ansipos++]=c;
-			if (c>=0x40&&c<=0x7e) { // final char
-				s->state=ST_NORM;
-				s->ansibuf[s->ansipos]=0;
-				switch (c) {
-					case '~': // 0x7e
-						if (s->ansipos==5&&s->ansibuf[2]=='1'&&s->ansibuf[3]=='1') // F1 - \e[11~
-							yascreen_pushch(s,YAS_K_F1);
-						if (s->ansipos==5&&s->ansibuf[2]=='1'&&s->ansibuf[3]=='2') // F2 - \e[12~
-							yascreen_pushch(s,YAS_K_F2);
-						if (s->ansipos==5&&s->ansibuf[2]=='1'&&s->ansibuf[3]=='3') // F3 - \e[13~
-							yascreen_pushch(s,YAS_K_F3);
-						if (s->ansipos==5&&s->ansibuf[2]=='1'&&s->ansibuf[3]=='4') // F4 - \e[14~
-							yascreen_pushch(s,YAS_K_F4);
-						if (s->ansipos==5&&s->ansibuf[2]=='1'&&s->ansibuf[3]=='5') // F5 - \e[15~
-							yascreen_pushch(s,YAS_K_F5);
-						if (s->ansipos==5&&s->ansibuf[2]=='1'&&s->ansibuf[3]=='7') // F6 - \e[17~
-							yascreen_pushch(s,YAS_K_F6);
-						if (s->ansipos==5&&s->ansibuf[2]=='1'&&s->ansibuf[3]=='8') // F7 - \e[18~
-							yascreen_pushch(s,YAS_K_F7);
-						if (s->ansipos==5&&s->ansibuf[2]=='1'&&s->ansibuf[3]=='9') // F8 - \e[19~
-							yascreen_pushch(s,YAS_K_F8);
-						if (s->ansipos==5&&s->ansibuf[2]=='2'&&s->ansibuf[3]=='1') // F10 - \e[21~
-							yascreen_pushch(s,YAS_K_F10);
-						if (s->ansipos==5&&s->ansibuf[2]=='2'&&s->ansibuf[3]=='3') // F11 - \e[23~
-							yascreen_pushch(s,YAS_K_F11);
-						if (s->ansipos==5&&s->ansibuf[2]=='2'&&s->ansibuf[3]=='4') // F12 - \e[24~
-							yascreen_pushch(s,YAS_K_F12);
-						if (s->ansipos==4&&s->ansibuf[2]=='2') // insert - \e[2~
-							yascreen_pushch(s,YAS_K_INS);
-						if (s->ansipos==4&&s->ansibuf[2]=='3') // delete - \e[3~
-							yascreen_pushch(s,YAS_K_DEL);
-						if (s->ansipos==4&&s->ansibuf[2]=='5') // pgup - \e[5~
-							yascreen_pushch(s,YAS_K_PGUP);
-						if (s->ansipos==4&&s->ansibuf[2]=='6') // pgdn - \e[6~
-							yascreen_pushch(s,YAS_K_PGDN);
-						if (s->ansipos==4&&(s->ansibuf[2]=='1'||s->ansibuf[2]=='7')) // home - \e[1~ \e[7~
-							yascreen_pushch(s,YAS_K_HOME);
-						if (s->ansipos==4&&(s->ansibuf[2]=='4'||s->ansibuf[2]=='8')) // end - \e[4~ \e[8~
-							yascreen_pushch(s,YAS_K_END);
-						break;
-					case 'R': { // \e[n;mR - cursor position report
-						int sx,sy;
-
-						sscanf((char *)s->ansibuf+2,"%d;%dR",&sy,&sx);
-						if (sx>10&&sy>3&&sx<=999&&sy<=999) { // ignore non-sane values
-							s->scrx=sx;
-							s->scry=sy;
-							s->haveansi=1;
-							yascreen_pushch(s,YAS_SCREEN_SIZE);
-						}
-						break;
-					}
-					case 'A': // ^up - \e[1;5A
-						if (s->ansipos==6&&s->ansibuf[2]=='1'&&s->ansibuf[3]==';'&&s->ansibuf[4]=='5')
-							yascreen_pushch(s,YAS_K_C_UP);
-						break;
-					case 'B': // ^down - \e[1;5B
-						if (s->ansipos==6&&s->ansibuf[2]=='1'&&s->ansibuf[3]==';'&&s->ansibuf[4]=='5')
-							yascreen_pushch(s,YAS_K_C_DOWN);
-						break;
-					case 'C': // ^right - \e[1;5C
-						if (s->ansipos==6&&s->ansibuf[2]=='1'&&s->ansibuf[3]==';'&&s->ansibuf[4]=='5')
-							yascreen_pushch(s,YAS_K_C_RIGHT);
-						break;
-					case 'D': // ^left - \e[1;5D
-						if (s->ansipos==6&&s->ansibuf[2]=='1'&&s->ansibuf[3]==';'&&s->ansibuf[4]=='5')
-							yascreen_pushch(s,YAS_K_C_LEFT);
-						break;
-				}
-			}
-			break;
-		case ST_ESC_O:
-			switch (c) {
-				case 'P': // F1 \eOP
-					yascreen_pushch(s,YAS_K_F1);
-					break;
-				case 'Q': // F2 \eOQ
-					yascreen_pushch(s,YAS_K_F2);
-					break;
-				case 'R': // F3 \eOR
-					yascreen_pushch(s,YAS_K_F3);
-					break;
-				case 'S': // F4 \eOS
-					yascreen_pushch(s,YAS_K_F4);
-					break;
-				case 'H': // home \eOH
-					yascreen_pushch(s,YAS_K_HOME);
-					break;
-				case 'F': // end \eOF
-					yascreen_pushch(s,YAS_K_END);
-					break;
-				case 'a': // ^up \eOa
-					yascreen_pushch(s,YAS_K_C_UP);
-					break;
-				case 'b': // ^down \eOb
-					yascreen_pushch(s,YAS_K_C_DOWN);
-					break;
-				case 'c': // ^right \eOc
-					yascreen_pushch(s,YAS_K_C_RIGHT);
-					break;
-				case 'd': // ^left \eOd
-					yascreen_pushch(s,YAS_K_C_LEFT);
-					break;
-			}
-			s->state=ST_NORM;
-			break;
-	}
-} // }}}
-
-inline int yascreen_getch_to(yascreen *s,int timeout) { // {{{
+static inline int yascreen_getch_to_gen(yascreen *s,int timeout,int key_none) { // {{{
 	int64_t toms=timeout*1000,tto;
 	struct timeval to,*pto=&to;
 	fd_set r;
 
-	memset(&r,0,sizeof r); // make clang static analyzer happier (llvm bug #8920)
 	if (!s)
-		return -1;
+		return key_none;
+
+	memset(&r,0,sizeof r); // make clang static analyzer happier (llvm bug #8920)
 
 	if (s->outcb) // we do not handle the input, so return immediately
 		timeout=-1;
@@ -1973,24 +1603,31 @@ inline int yascreen_getch_to(yascreen *s,int timeout) { // {{{
 			return key;
 		}
 		if (s->outcb)
-			return -1;
+			return key_none;
 		if (STDOUT_FILENO<0)
-			return -1;
+			return key_none;
 		FD_ZERO(&r);
 		FD_SET(STDOUT_FILENO,&r);
 		if (-1!=select(STDOUT_FILENO+1,&r,NULL,NULL,pto)) {
 			unsigned char c; // important to be unsigned, so codes>127 do not expand as negative int values
 
 			if (FD_ISSET(STDOUT_FILENO,&r)&&sizeof c==read(STDOUT_FILENO,&c,sizeof c)) {
+				#if YASCREEN_VERSIONED
+				if (key_none==YAS_K_NONE) // default behaviour, new symbols
+					yascreen_feed_193(s,c);
+				else
+					yascreen_feed_179(s,c);
+				#else
 				yascreen_feed(s,c);
+				#endif
 				continue; // check if feed has yielded a key
 			}
 		}
 		if (pto&&(timeout>0||s->escto)&&to.tv_sec==0&&to.tv_usec==0) { // return because of timeout
 			if (timeout<0) // nowait is set
-				return -1;
+				return key_none;
 			if (!toms&&timeout>0) // timeout is finished
-				return -1;
+				return key_none;
 			if (!toms)
 				tto=s->escto;
 			else
@@ -2003,6 +1640,11 @@ inline int yascreen_getch_to(yascreen *s,int timeout) { // {{{
 	}
 } // }}}
 
+symver_d(yascreen_getch_to_193,yascreen_getch_to,YASCREEN_1.93) // {{{
+inline int V(yascreen_getch_to,V193)(yascreen *s,int timeout) {
+	return yascreen_getch_to_gen(s,timeout,YAS_K_NONE);
+} // }}}
+
 inline void yascreen_ungetch(yascreen *s,int key) { // {{{
 	int *tk;
 
@@ -2050,10 +1692,11 @@ inline void yascreen_esc_to(yascreen *s,int timeout) { // {{{
 	s->escto=(timeout>=0)?timeout:YAS_DEFAULT_ESCTO;
 } // }}}
 
-inline int yascreen_peekch(yascreen *s) { // {{{
+symver_d(yascreen_peekch_193,yascreen_peekch,YASCREEN_1.93) // {{{
+inline int V(yascreen_peekch,V193)(yascreen *s) {
 	int ch=yascreen_getch_nowait(s);
 
-	if (ch!=-1)
+	if (ch!=YAS_K_NONE)
 		yascreen_ungetch(s,ch);
 	return ch;
 } // }}}
@@ -2093,3 +1736,245 @@ inline void yascreen_line_flush(yascreen *s,int on) { // {{{
 	s->lineflush=!!on;
 } // }}}
 
+inline wchar_t yascreen_getwch_to(yascreen *s,int timeout) { // {{{
+	int ch;
+
+	if (!s)
+		return YAS_K_NONE;
+	if (!s->isunicode)
+		return YAS_K_NONE;
+
+	ch=yascreen_getch_to(s,timeout);
+
+	if (YAS_IS_CC(ch))
+		return ch;
+	if (ch>=0&&ch<=0x7f)
+		return ch;
+	if ((ch&0xe0)==0xc0) { // 2 byte seq
+		// unicode mode guarantees that we have a valid sequence
+		unsigned char c2=yascreen_getch_nowait(s);
+		wchar_t w;
+
+		w=((ch&0x1f)<<6)|(c2&0x3f);
+		return w;
+	}
+	if ((ch&0xf0)==0xe0) { // 3 byte seq
+		// unicode mode guarantees that we have a valid sequence
+		unsigned char c2=yascreen_getch_nowait(s);
+		unsigned char c3=yascreen_getch_nowait(s);
+		wchar_t w;
+
+		w=((ch&0x0f)<<12)|((c2&0x3f)<<6)|(c3&0x3f);
+		return w;
+	}
+	if ((ch&0xf8)==0xf0) { // 4 byte seq
+		// unicode mode guarantees that we have a valid sequence
+		unsigned char c2=yascreen_getch_nowait(s);
+		unsigned char c3=yascreen_getch_nowait(s);
+		unsigned char c4=yascreen_getch_nowait(s);
+		wchar_t w;
+
+		w=((ch&0x07)<<18)|((c2&0x3f)<<12)|((c3&0x3f)<<6)|(c4&0x3f);
+		return w;
+	}
+	//if ((ch&0xfc)==0xf8) { // 5 byte seq
+	//}
+	//if ((str[i]&0xfe)==0xfc) { // 6 byte seq
+	//}
+	return YAS_K_NONE;
+} // }}}
+
+inline void yascreen_ungetwch(yascreen *s,wchar_t key) { // {{{
+	if (!s)
+		return;
+	if (!s->isunicode)
+		return;
+
+	if (YAS_IS_CC(key))
+		yascreen_ungetch(s,key);
+	else {
+		char ns[MB_CUR_MAX+1];
+		int rc=wctomb(ns,key);
+		int i;
+
+		if (rc<=0)
+			return;
+		for (i=rc-1;i>0;i--)
+			yascreen_ungetch(s,ns[i]);
+	}
+} // }}}
+
+inline wchar_t yascreen_peekwch(yascreen *s) { // {{{
+	wchar_t ch;
+
+	if (!s)
+		return YAS_K_NONE;
+	if (!s->isunicode)
+		return YAS_K_NONE;
+
+	ch=yascreen_getwch_nowait(s);
+
+	if (ch!=YAS_K_NONE)
+		yascreen_ungetwch(s,ch);
+	return ch;
+} // }}}
+
+// get yascreen_feed with the new symbol or unversioned
+#include "yascreen_feed.c"
+
+#if YASCREEN_VERSIONED // redefine YAS_K_* {{{
+
+#undef YAS_K_ALT
+#define YAS_K_ALT(code) (((code)&0xff)|0x200)
+
+#define YAS_K_NONE -1 // no key available
+// extended keys send as escape sequences
+// function keys with ALT/CTRL/SHIFT
+#define YAS_K_F1 0x100
+#define YAS_K_F2 0x101
+#define YAS_K_F3 0x102
+#define YAS_K_F4 0x103
+#define YAS_K_F5 0x104
+#define YAS_K_F6 0x105
+#define YAS_K_F7 0x106
+#define YAS_K_F8 0x107
+#define YAS_K_F9 0x108
+#define YAS_K_F10 0x109
+#define YAS_K_F11 0x10a
+#define YAS_K_F12 0x10b
+#define YAS_K_S_F1 0x10c
+#define YAS_K_S_F2 0x10d
+#define YAS_K_S_F3 0x10e
+#define YAS_K_S_F4 0x10f
+#define YAS_K_S_F5 0x110
+#define YAS_K_S_F6 0x111
+#define YAS_K_S_F7 0x112
+#define YAS_K_S_F8 0x113
+#define YAS_K_S_F9 0x114
+#define YAS_K_S_F10 0x115
+#define YAS_K_S_F11 0x116
+#define YAS_K_S_F12 0x117
+#define YAS_K_C_F1 0x118
+#define YAS_K_C_F2 0x119
+#define YAS_K_C_F3 0x11a
+#define YAS_K_C_F4 0x11b
+#define YAS_K_C_F5 0x11c
+#define YAS_K_C_F6 0x11d
+#define YAS_K_C_F7 0x11e
+#define YAS_K_C_F8 0x11f
+#define YAS_K_C_F9 0x120
+#define YAS_K_C_F10 0x121
+#define YAS_K_C_F11 0x122
+#define YAS_K_C_F12 0x123
+#define YAS_K_A_F1 0x124
+#define YAS_K_A_F2 0x125
+#define YAS_K_A_F3 0x126
+#define YAS_K_A_F4 0x127
+#define YAS_K_A_F5 0x128
+#define YAS_K_A_F6 0x129
+#define YAS_K_A_F7 0x12a
+#define YAS_K_A_F8 0x12b
+#define YAS_K_A_F9 0x12c
+#define YAS_K_A_F10 0x12d
+#define YAS_K_A_F11 0x12e
+#define YAS_K_A_F12 0x12f
+#define YAS_K_LEFT 0x130
+#define YAS_K_UP 0x131
+#define YAS_K_DOWN 0x132
+#define YAS_K_RIGHT 0x133
+#define YAS_K_HOME 0x134
+#define YAS_K_END 0x135
+#define YAS_K_PGUP 0x136
+#define YAS_K_PGDN 0x137
+#define YAS_K_INS 0x138
+#define YAS_K_DEL 0x139
+#define YAS_K_C_LEFT 0x13a
+#define YAS_K_C_UP 0x13b
+#define YAS_K_C_DOWN 0x13c
+#define YAS_K_C_RIGHT 0x13d
+// ALT+letter
+#define YAS_K_A_BT YAS_K_ALT('`')
+#define YAS_K_A_1 YAS_K_ALT('1')
+#define YAS_K_A_2 YAS_K_ALT('2')
+#define YAS_K_A_3 YAS_K_ALT('3')
+#define YAS_K_A_4 YAS_K_ALT('4')
+#define YAS_K_A_5 YAS_K_ALT('5')
+#define YAS_K_A_6 YAS_K_ALT('6')
+#define YAS_K_A_7 YAS_K_ALT('7')
+#define YAS_K_A_8 YAS_K_ALT('8')
+#define YAS_K_A_9 YAS_K_ALT('9')
+#define YAS_K_A_0 YAS_K_ALT('0')
+#define YAS_K_A_MINUS YAS_K_ALT('-')
+#define YAS_K_A_EQ YAS_K_ALT(' ')
+#define YAS_K_A_BSP YAS_K_ALT(0x7f)
+#define YAS_K_A_TLD YAS_K_ALT('~')
+#define YAS_K_A_EXCL YAS_K_ALT('!')
+#define YAS_K_A_AT YAS_K_ALT('@')
+#define YAS_K_A_HASH YAS_K_ALT('#')
+#define YAS_K_A_POUND YAS_K_ALT('$')
+#define YAS_K_A_PERC YAS_K_ALT('%')
+#define YAS_K_A_CARRET YAS_K_ALT('^')
+#define YAS_K_A_AND YAS_K_ALT('&')
+#define YAS_K_A_STAR YAS_K_ALT('*')
+#define YAS_K_A_OBRA YAS_K_ALT('(')
+#define YAS_K_A_CBRA YAS_K_ALT(')')
+#define YAS_K_A_UND YAS_K_ALT('_')
+#define YAS_K_A_PLUS YAS_K_ALT('+')
+#define YAS_K_A_a YAS_K_ALT('a')
+#define YAS_K_A_b YAS_K_ALT('b')
+#define YAS_K_A_c YAS_K_ALT('c')
+#define YAS_K_A_d YAS_K_ALT('d')
+#define YAS_K_A_e YAS_K_ALT('e')
+#define YAS_K_A_f YAS_K_ALT('f')
+#define YAS_K_A_g YAS_K_ALT('g')
+#define YAS_K_A_h YAS_K_ALT('h')
+#define YAS_K_A_i YAS_K_ALT('i')
+#define YAS_K_A_j YAS_K_ALT('j')
+#define YAS_K_A_k YAS_K_ALT('k')
+#define YAS_K_A_l YAS_K_ALT('l')
+#define YAS_K_A_m YAS_K_ALT('m')
+#define YAS_K_A_n YAS_K_ALT('n')
+#define YAS_K_A_o YAS_K_ALT('o')
+#define YAS_K_A_p YAS_K_ALT('p')
+#define YAS_K_A_q YAS_K_ALT('q')
+#define YAS_K_A_r YAS_K_ALT('r')
+#define YAS_K_A_s YAS_K_ALT('s')
+#define YAS_K_A_t YAS_K_ALT('t')
+#define YAS_K_A_u YAS_K_ALT('u')
+#define YAS_K_A_v YAS_K_ALT('v')
+#define YAS_K_A_w YAS_K_ALT('w')
+#define YAS_K_A_x YAS_K_ALT('x')
+#define YAS_K_A_y YAS_K_ALT('y')
+#define YAS_K_A_z YAS_K_ALT('z')
+#define YAS_SCREEN_SIZE 0x800
+#define YAS_TELNET_SIZE 0x801
+
+// }}}
+
+// {{{
+
+symver_o(yascreen_getch_to_179,yascreen_getch_to,YASCREEN_1.79) // {{{
+inline int yascreen_getch_to_179(yascreen *s,int timeout) {
+	return yascreen_getch_to_gen(s,timeout,YAS_K_NONE);
+} // }}}
+
+symver_o(yascreen_peekch_179,yascreen_peekch,YASCREEN_1.79) // {{{
+inline int yascreen_peekch_179(yascreen *s) {
+	int ch=yascreen_getch_to_179(s,-1);
+
+	if (ch!=YAS_K_NONE)
+		yascreen_ungetch(s,ch);
+	return ch;
+} // }}}
+
+// now get yascreen_feed old version
+
+#undef symver_V
+#define symver_V(impl,sym,ver) symver_o(impl,sym,ver)
+#undef V193
+#define V193 _179
+#undef YASCREEN_193
+#define YASCREEN_193 YASCREEN_1.79
+#include "yascreen_feed.c"
+
+#endif
